/*----------------------------------------------------------------------------------------
|
| 	[GEIGER COUNTER	with ATMega88]
| 	Copyright (C) [2012]  [Clemens Verstappen DL3ETW] 
| 	 
|  	This program is free software; you can redistribute it and/or modify it under the terms  
| 	of the GNU General Public License as published by the Free Software Foundation; either 
|	version 3 of the License, or (at your option) any later version.	 
|	This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
|  	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
|	See the GNU General Public License for more details. 
|	You should have received a copy of the GNU General Public License along with this program;
| 	if not, see <http://www.gnu.org/licenses/>. 
| 
| 	Hardware: 	ATMega88 with LCD 2*8 in one line, plus switch for display toggle, 
|				TTL pulses from geiger counter tube
|   Funktions: 	counts, averages and displays radioactive desintegrations from counter tube
| 				in CPM (counts per minute) and nSv/h /nano Sievert per hour.
| 				Mmonitors battery and high voltage with AD converter and displays values. 
|   			Swiches power off if battery <3.5V and after 10 minutes.
| 	Fusebits:
|	Ext. Crystal Osc. 20  MHz; Start-up time PWRDWN/RESET: 16K CK/14 CK + 65 ms
|	use 20MHz x-tal.
|---------------------------------------------------------------------------------------- */

#define 	F_CPU 20000000UL		
#include 	"lcd.h"				
#include 	<avr/interrupt.h>
#include 	<C:\WinAVR-20100110\avr/include/stdlib.h>

#include 	<util/delay.h>	

char   		zeile1[16], zeile2[16];
uint8_t 	display_mode, ad_channel=0,counter=0,conversion_counter=0;
uint8_t 	average_array[15];
uint16_t   	temp, CPM=0, batterie_voltage=0, high_voltage=0, ADC_results[8];
uint16_t 	auto_off_counter=0;

volatile struct BitVariablen
{											
	int ADC0_update 	:1;
	int ADC1_update 	:1;
	int	realtime_update :1;
	int power_on		:1;					// bit for P-channel mosfet gate 
}status = {1};

SIGNAL(INT0_vect) 							// RADIATION-COUNTER TUBE 
	{										// external  Interrupt0, falling edge  
	PORTB |= (1<<PORTB2);					// switches pulse on for speaker
	TCNT0 =	0b00000000;						// Timer register clear
	counter++; 								// how many clicks per second?
	}

SIGNAL(TIMER0_COMPA_vect) 					// SPEAKER PULSE LENGHT 
	{										// compare match after 200us switches speaker off  
	PORTB &= (~(1<<PORTB2));				// switch pulse off for speaker
	} 

SIGNAL(INT1_vect) 							// DISPLAY MODE SWITCH
	{										// external interrupt from switch 
	display_mode++;							// toggle between display modes
	if 	(display_mode==4)
	display_mode=0;
	}
								
SIGNAL(TIMER1_COMPA_vect) 					// MAIN INTERVAL, start for main loop
	{										// Signal(Timer_Compa_vect) Interrupt each second
	PORTB ^= (1<<PORTB1);					// LED toggle 0.5 Hz
	status.realtime_update = 1;				// indicate new calculation in main program
	} 

SIGNAL(ADC_vect) 							// AD CONVERTER 
	{
	ADCSRA	=0b10000100;					// Conversion stop, interrupt stop
	ADC_results[ad_channel] = ADC;			// save data in array
	conversion_counter++;					// use second conversion only, skip first one
	}

void power_off(void)
	{
	lcd_gotoxy(0,0);					// 	address LCD line 0
	lcd_puts("LOW BATT");
	lcd_gotoxy(0,1);					// 	
	lcd_puts("        ");
	_delay_ms(5000);
	PORTC &= (~(1<<PORTC2));			// switch power off, portc2 to low
	cli();								// here stop, 	
	for(;;);							// endless loop
	}

void auto_power_off(void)
	{
	lcd_gotoxy(0,0);					// 	address LCD line 0
	lcd_puts("  Power ");
	lcd_gotoxy(0,1);					// 	
	lcd_puts("off     ");
	_delay_ms(5000);
	PORTC &= (~(1<<PORTC2));			// switch power off, portc2 to low
	cli();								// here stop, 	
	for(;;);							// endless loop
	}

void load_array(void)						// load initial values fr the array
	{
	uint8_t i;
	for(i = 0; i <15; i++)
	average_array[i]=1;
	}

void shift_array(void)						// shift all elements in array by one
	{										// last one ist lost
	uint8_t i;
	for(i = 0; i <15; i++)
	average_array[i] = average_array[i+1];
	}

uint16_t mean_average(void)					// mean value of array
	{
	uint16_t summe=0;
	uint8_t i;
	for(i = 0; i < 15; i++)
	summe += average_array[i];
	return summe; 
	}

void display_lcd(uint8_t display_mode)
	{
	uint8_t i;

	if (display_mode==0)					// 	CPM, clicks per minute					
		{
		itoa(CPM, zeile2, 10);
		lcd_gotoxy(0,0);					// 	address LCD line 0
		lcd_puts("   CPM :");
		lcd_gotoxy(0,1);
		lcd_puts("        ");
		lcd_gotoxy(0,1);
		for(i = 0; zeile2[i] && i < 8; i++)	//
		lcd_putc(zeile2[i]);
		}

	if (display_mode==1)					//	DOSE EQUIVALENT in uSv/h
		{
		itoa(CPM/9*50, zeile2, 10);			//	calibration value 1000*CPM/60*0.33
		lcd_gotoxy(0,0);					// 	address LCD line 0
		lcd_puts(" nSv/h :");
		lcd_gotoxy(0,1);
		lcd_puts("        ");
		lcd_gotoxy(0,1);
		for(i = 0; zeile2[i] && i < 8; i++)	// 	
		lcd_putc(zeile2[i]);
		}

	if (display_mode==2)					// 	BATTERY VOLTAGE / mV
		{
		itoa(batterie_voltage, zeile2, 10);
		lcd_gotoxy(0,0);					// 	address LCD line 0
		lcd_puts("Batt/mV:");
		lcd_gotoxy(0,1);
		lcd_puts("        ");
		lcd_gotoxy(0,1);
		for(i = 0; zeile2[i] && i < 8; i++)	// 	
		lcd_putc(zeile2[i]);
		}
	
	if (display_mode==3)					//	HIGH VOLTAGE / V from tube
		{
		itoa(high_voltage, zeile2, 10);
		lcd_gotoxy(0,0);					// 	address LCD line 0
		lcd_puts("  HV/V :");
		lcd_gotoxy(0,1);
		lcd_puts("        ");
		lcd_gotoxy(0,1);
		for(i = 0; zeile2[i] && i < 8; i++)	// 	
		lcd_putc(zeile2[i]);
		}
	}		

int main(void) 	
	{
	 
	DDRB 	= 0b00001110;			// PB0 LCD RS, PB1 LED,PB2 for speaker, PB3 for meter
	DDRC 	= 0b00000100;			// PortC all input, except PC2 - power_on bit	
	PORTC	= 0b00000100;			// initiallize to null, switch power_on with high

// TIMER 0							//	determinds length of click for speaker

	TCCR0A	= 0b00000010;			// timer 0, init,wgm00=0  wgm01=1 CTC, wgm02=0
	TCCR0B	= 0b00000101;			// CLK/1024 = 51us per clk
	OCR0A	= 0b00000100;			// output compare at 4 *51us = 200 us pulslength speaker
	TIMSK0	= 0b00000010;			// output compare match A interrupt enable

// TIMER 1							// starts main loop every second
	
	OCR1A	= 0x4C4A;				// timer 1, compare match register load, 4C4B = 1s ;2626 1313 = 250ms
	TCCR1B	= 0b00001101;			// set WGM12, clear on compare match, divide by 1024
	TIMSK1	= 0b00000010;			// timer1 compare match 1 interrupt enable

// TIMER 2							// PWM for analog meter (100uA)	

	TCCR2A	= 0b10000011;			// WGM21 and WGM20 for fast PWM mode
	TCCR2B	= 0b00000111;			// prescaler to /1024 / =19.5 KHz, PWM frequency/256 = 76Hz
	OCR2A	= 0b00000111;			// initialize with 3

	EICRA	|= (1<<ISC01);			// falling edge for external interrupt 0, radiation counter tube
	EIMSK	|= (1<<INT0);			// enable external interrupt 0	
  	
	EICRA	|= (1<<ISC11)|(1<<ISC10);		// rising edge for external interrupt 1, toggle switch
	EIMSK	|= (1<<INT1);					// enable external interrupt 1
	
//  initialize A/D

	ADMUX	=0b11000000;			// 1100 - use internal 1.1V referenz, 
									// 0000 - ADC0, 0001 ADC1 A/D initialisieren, 
	ADCSRA	=0b11101100;			// 1100 / Bit 6 start conversion, 
									// 0100	/ clk / 16, Bit 3 is interrup enable	

	lcd_init(LCD_DISP_ON);			// initialize LCD
 	void lcd_clrscr();				// clear LCD

	void load_array();				// Average array load with default values

	sei ();							// nun Interrupts freigeben

	for (;;)  						// endless loop
		{
		if (status.realtime_update) 			// loop each second	
			{								
			auto_off_counter++;
			if (auto_off_counter>=600)			// auto power off after 10 minutes (600s)
				auto_power_off();

			shift_array();						// clear a14
			average_array[14]=counter;			// load with counts of past 15 seconds
			counter =0x00;						// reset counter
			CPM =  mean_average()*4;			// estimate mean value and times 4 for one min.

			temp = CPM;							// copy into temp,
			OCR2A	= (temp >> 4); 				// set meter value, shift 8x right
												// to fit into 8bit. about 4000cpm for full scale
			status.realtime_update = 0;

			if (conversion_counter==3)			// skip the first AD conversions
				{
				batterie_voltage = ADC_results[0]*67/13;	//calibration_value_batt;	//75/13
				high_voltage = ADC_results[1];	//calibration_value_hv
				if (batterie_voltage<3500)		// switch off, if accu is empty
					power_off();
				conversion_counter=0;			
				display_lcd(display_mode);
				ad_channel++;
				if (ad_channel == 2)			// only 2 AD converters in a row
				ad_channel = 0; 
				}
									
			ADMUX	=0b11000000+ad_channel;		//	select 0ADC1
			ADCSRA	=0b11001100;				//  AD start and interrupt enable		
			}
		}
}


	// Pin asignment 
	// Port B Bit 0 PB0 (PCINT0/CKL0/ICP1)		LCD RS									
	//        Bit 1 PB1 (OC1A/PCINT1)			LED 
	//        Bit 2 PB2	(SS/OC1B/PCINT3)		speaker
  	//        Bit 3 PB3	(MOSI/OC2A/PCINT3))		Meter analog 100uA, 	(ISP MOSI)
  	//        Bit 4 PB4	(MISO/PCINT4))					(ISP MISO)
  	//        Bit 5 PB5 (SCK/PCINT5)					(ISP SCK)
  	//        Bit 6 PB6	(PCINT6/XTAL1/TOSC1)	20MHz X-TAL
	//		  Bit 7 PB7	(PCINT7/XTAL2/TOSC2)	20MHz X-TAL

 	// Port C Bit 0 PC0 (ADC0/PCINT8)			battery voltage, AD 
  	//        Bit 1 PC1	(ADC1/PCINT9)			high voltage, AD from turadiation counter tube
 	//        Bit 2 PC2	(ADC2/PCINT10)			power_off, output to high side switch	
  	//        Bit 3 PC3	(ADC3/PCINT11)			
  	//        Bit 4 PC4	(ADC4/SDA/PCINT12)		 
  	//        Bit 5 PC5	(ADC5/SCL/PCINT13)		
  	//        Bit 6 PC6	(PCINT/RESET)			Reset
  			          										
	// Port D Bit 0 PD0	(PCINT16/RXD)			LCD E			
	//        Bit 1 PD1	(PCINT17/TXD)			LCD R/W
  	//        Bit 2 PD2	(PCINT18/INT0)			radiation counter tube
  	//        Bit 3 PD3	(PCINT19/OC2B/INT1)		display mode switch
  	//        Bit 4 PD4	(PCINT20/XCK/T0)		LCD Data dB4 	
  	//        Bit 5 PD5	(PCINT21/OC0B/T1)		LCD Data dB5 	
  	//        Bit 6 PD6	(PCINT22/OC0A/AIN0)		LCD Data dB6 	
  	//        Bit 7 PD7	(PCINT23/AIN1)			LCD Data dB7 			
